import React, { useEffect, useRef, useState } from 'react';
import {
  DeleteOutlined,
  InboxOutlined,
  LoadingOutlined,
  PaperClipOutlined,
  WarningOutlined,
} from '@ant-design/icons';
import type { UploadProps } from 'antd';
import { Upload, Progress, Button, Tooltip } from 'antd';
import { PageContainer, ProCard } from '@ant-design/pro-components';
import type { RcFile } from 'antd/lib/upload';
import type { UploadFileStatus } from 'antd/lib/upload/interface';
import SparkMD5 from 'spark-md5';

const { Dragger } = Upload;

const renderSize = (filesize: any) => {
  const resize = Number(filesize);
  if (resize == 0) {
    return '0 Bytes';
  }
  const unitArr = ['Bytes', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'];
  let index = 0;
  index = Math.floor(Math.log(resize) / Math.log(1024));
  const size = resize / Math.pow(1024, index);
  if (index > 0) {
    return size.toFixed(0) + unitArr[index];
  } else {
    return size + unitArr[index];
  }
};

type WorkStatus = 'wait' | 'march' | 'finish';
type WorkData = {
  uid: string;
  name: string;
  status: UploadFileStatus;
  md5?: string;
  size?: number;
};

const CalculateMD5: React.FC = () => {
  // 可以用来终止计算文件 MD5
  const preStatusRef = useRef(false);
  // 上传任务状态
  const workStatusList = useRef<
    {
      uid: string;
      status: WorkStatus;
      preProgress: string;
      data: WorkData;
      file: RcFile;
    }[]
  >([]);
  // 上传任务列表
  const [workList, setWorkList] = useState<
    {
      uid: string;
      data: WorkData;
    }[]
  >([]);
  useEffect(() => {
    if (workStatusList.current.length == 0) return;
    for (let i = 0; i < workStatusList.current.length; i++) {
      // 有正在计算的，就跳过新任务
      if (workStatusList.current[i].status == 'march') return;
    }
    for (let i = 0; i < workStatusList.current.length; i++) {
      // 有正在计算的，就跳过新任务
      if (workStatusList.current[i].status == 'wait') {
        workStatusList.current[i].status = 'march';
        work(workStatusList.current[i].file);
        return;
      }
    }
  }, [workList]);

  // 重置
  const reset = () => {
    preStatusRef.current = false;
    workStatusList.current = [];
    setWorkList([]);
  };

  const props: UploadProps = {
    accept: '.mp4,video/mp4,.mov,video/mov',
    multiple: true,
    fileList: workList.map((it) => it.data),
    itemRender: (originNode, file, fileList, actions) => {
      let fileNode;
      if (file.status == 'uploading')
        fileNode = (
          <>
            <LoadingOutlined /> {file.name}
          </>
        );
      else if (file.status == 'done')
        fileNode = (
          <>
            <PaperClipOutlined /> {file.name}
          </>
        );
      else
        fileNode = (
          <>
            <WarningOutlined /> {file.name}
          </>
        );

      const tmpWork = workStatusList.current.filter((it) => {
        return it.uid == file.uid;
      })[0];
      if (!tmpWork) return null;

      return (
        <ProCard
          title={
            <>
              {fileNode} {renderSize(file.size || '')}
            </>
          }
          extra={
            <Tooltip title="search">
              <Button
                type="dashed"
                shape="circle"
                icon={<DeleteOutlined />}
                onClick={actions.remove}
              />
            </Tooltip>
          }
        >
          <div>MD5: {tmpWork.data.md5 && <>{tmpWork.data.md5}</>}</div>
          <div style={{ width: '80%' }}>
            <Progress percent={Number(tmpWork.preProgress)} />
          </div>
        </ProCard>
      );
    },
    onRemove: (file) => {
      workStatusList.current = workStatusList.current.filter((it) => {
        return it.uid != file.uid;
      });
      setWorkList(
        workStatusList.current.map((it) => {
          return { uid: it.uid, data: it.data };
        }),
      );
    },
    beforeUpload: (file, FileList) => {
      const list = FileList.filter((it) => {
        return !!it.uid;
      });
      if (list.length != FileList.length) return;
      workStatusList.current = [
        ...workStatusList.current,
        ...FileList.map((it) => {
          return {
            uid: it.uid,
            status: 'wait' as WorkStatus,
            preProgress: '0',
            data: {
              uid: it.uid,
              name: it.name,
              size: it.size,
              status: 'uploading' as UploadFileStatus,
            },
            file: it,
          };
        }),
      ];
      setWorkList(
        workStatusList.current.map((it) => {
          return { uid: it.uid, data: it.data };
        }),
      );
      return false;
    },
  };

  function work(file: RcFile) {
    preStatusRef.current = true;
    // 上传文件所有分片的 md5
    const params: {
      uid: string;
      file: Record<string, any>;
      chunks: Record<string, any>[];
    } = {
      uid: file.uid,
      file: {},
      chunks: [],
    };
    params.file.fileName = file.name;
    params.file.fileSize = file.size;

    // 分片每片大小 5M
    const chunkSize = 1024 * 1024 * 5;
    // 分片数量
    const chunks = Math.ceil(file.size / chunkSize);
    // 当前分片序号
    let currentChunk = 0;

    const sparkAll = new SparkMD5.ArrayBuffer();
    const spark = new SparkMD5.ArrayBuffer();

    // 分片的 Md5
    const chunkFileReader = new FileReader();
    chunkFileReader.onerror = function () {
      console.warn('oops, something went wrong.');
    };
    chunkFileReader.onload = function (e) {
      sparkAll.append(e.target!!.result);
      spark.append(e.target!!.result);
      params.chunks.push({
        chunk: currentChunk + 1,
        start: currentChunk * chunkSize, // 计算分片的起始位置
        end:
          currentChunk * chunkSize + chunkSize >= file.size
            ? file.size
            : currentChunk * chunkSize + chunkSize, // 计算分片的结束位置
        chunkMd5: spark.end(null),
        chunks,
        currentBuffer: e.target!!.result,
      });

      currentChunk++; // 累加分片

      // 文件删除了，就不要在继续计算了
      const d = workStatusList.current.filter((it) => {
        return it.uid == file.uid;
      });
      if (d.length == 0) return;
      if (!preStatusRef.current) return;

      if (currentChunk < chunks) {
        workStatusList.current = workStatusList.current.map((it) => {
          if (it.uid == file.uid)
            it.preProgress = ((currentChunk / chunks) * 100).toFixed(2);
          return it;
        });
        setWorkList([
          ...workStatusList.current.map((it) => {
            return { uid: it.uid, data: it.data };
          }),
        ]);
        loadNext();
      } else {
        params.file.fileChunks = params.chunks.length;
        params.file.fileMd5 = sparkAll.end(null);
        workStatusList.current = workStatusList.current.map((it) => {
          if (it.uid == file.uid) {
            it.status = 'finish';
            it.preProgress = '100';
            it.data.status = 'done';
            it.data.md5 = params.file.fileMd5;
          }
          return it;
        });
        setWorkList([
          ...workStatusList.current.map((it) => {
            return { uid: it.uid, data: it.data };
          }),
        ]);
      }
    };

    // 兼容性的处理
    const blobSlice =
      File.prototype.slice ||
      File.prototype.mozSlice ||
      File.prototype.webkitSlice;

    function loadNext() {
      const start = currentChunk * chunkSize;
      const end =
        start + chunkSize >= file.size ? file.size : start + chunkSize;
      chunkFileReader.readAsArrayBuffer(blobSlice.call(file, start, end));
    }

    loadNext();
  }

  return (
    <PageContainer>
      <ProCard
        actions={[
          <Button
            key="reset"
            type="link"
            icon={<DeleteOutlined />}
            onClick={reset}
            block
          >
            清空
          </Button>,
        ]}
      >
        <Dragger {...props}>
          <p className="ant-upload-drag-icon">
            <InboxOutlined />
          </p>
          <p className="ant-upload-text">单击或将文件拖动到此区域可以计算MD5</p>
          <p className="ant-upload-hint">
            文件大小对计算速度有显著影响，请耐心等待计算完成。
          </p>
        </Dragger>
      </ProCard>
    </PageContainer>
  );
};

export default CalculateMD5;
